import itertools
import json
from monty.io import zopen
import numpy as np


class Atom(object):
    """Represents a single persistent atom during a simulation.

    Attributes:
        index (int): Unique numeric index identifying this atom.
        in_site (int): Site index for the site this atom
            currently occupies.
        frac_coords (np.array): Numpy array containing the current fractional
            coordinates for this atom.
        trajectory (list): List of site indices occupied at each timestep.

    """

    def __init__(self, index, species_string=None):
        """Initialise an Atom object.

        Args:
            index (int): Numerical index for this atom. Used to identify this atom
                in analysed structures.

        Returns:
            None

        """
        self.index = index
        self.in_site = None
        self._frac_coords = None
        self.trajectory = []

    def __str__(self):
        """Return a string representation of this atom.

        Args:
            None

        Returns:
            (str)

        """
        string = f"Atom: {self.index}"
        return string

    def __repr__(self):
        string = (
            "site_analysis.Atom("
            f"index={self.index}, "
            f"in_site={self.in_site}, "
            f"frac_coords={self._frac_coords})"
        )
        return string

    def reset(self):
        """Reset the state of this Atom.

        Clears the `in_site` and `trajectory` attributes.

        Returns:
            None

        """
        self.in_site = None
        self._frac_coords = None
        self.trajectory = []

    def assign_coords(self, structure):
        """Assign fractional coordinates to this atom from a 
        pymatgen Structure.

        Args:
            structure (pymatgen.Structure): The Structure to use for this atom's
                fractional coordinates.

        Returns:
            None

        """
        self._frac_coords = structure[self.index].frac_coords

    @property
    def frac_coords(self):
        """Getter for the fractional coordinates of this atom.

        Raises:
            AttributeError: if the fractional coordinates for this atom have
                not been set.

        """
        if self._frac_coords is None:
            raise AttributeError("Coordinates not set for atom {}".format(self.index))
        else:
            return self._frac_coords

    def as_dict(self):
        d = {
            "index": self.index,
            "in_site": self.in_site,
            "frac_coords": self._frac_coords.tolist(),
        }
        return d

    @classmethod
    def from_dict(cls, d):
        atom = cls(index=d["index"])
        atom.in_site = d["in_site"]
        atom._frac_coords = np.array(d["frac_coords"])
        return atom

    def to(self, filename=None):
        s = json.dumps(self.as_dict())
        if filename:
            with zopen(filename, "wt") as f:
                f.write("{}".format(s))
        return s

    @classmethod
    def from_str(cls, input_string):
        d = json.loads(input_string)
        return cls.from_dict(d)

    @classmethod
    def from_file(cls, filename):
        with zopen(filename, "rt") as f:
            contents = f.read()
        return cls.from_str(contents)


def atoms_from_species_string(structure, species_string):
    atoms = [
        Atom(index=i)
        for i, s in enumerate(structure)
        if s.species_string == species_string
    ]
    return atoms


def atoms_from_indices(indices):
    return [Atom(index=i) for i in indices]



from .site import Site

class VoronoiSite(Site):
    """Site subclass corresponding to Voronoi cells centered
    on fixed positions.

    Attributes:
        frac_coords (np.array): Fractional coordinates of the site centre.

    """
    
    def __init__(self, frac_coords, label=None):
        """Create a ``VoronoiSite`` instance.
        
        Args:
            frac_coords (np.array): Fractional coordinates of the site centre.
            label (:str:, optional): Optional label for this site. Default is `None`.

        Returns:
            None

        """
        super(VoronoiSite, self).__init__(label=label)
        self.frac_coords = frac_coords

    def as_dict(self):
        """Json-serializable dict representation of this VoronoiSite.

        Args:
            None

        Returns:
            (dict)

        """
        d = super(VoronoiSite, self).as_dict()
        d['frac_coords'] = self.frac_coords
        return d

    @classmethod
    def from_dict(cls, d):
        """Create a VoronoiSite object from a dict representation.

        Args:
            d (dict): The dict representation of this Site.

        Returns:
            (VoronoiSite)

        """
        voronoi_site = cls( frac_coords=d['frac_coords'] )
        voronoi_site.label = d.get('label')
        return voronoi_site 




from collections import Counter
from tqdm import tqdm, tqdm_notebook
from .polyhedral_site_collection import PolyhedralSiteCollection
from .polyhedral_site import PolyhedralSite
from .voronoi_site import VoronoiSite
from .voronoi_site_collection import VoronoiSiteCollection
from .spherical_site import SphericalSite
from .spherical_site_collection import SphericalSiteCollection

class Trajectory(object):
    """Class for performing sites analysis on simulation trajectories."""

    site_collection_types = { PolyhedralSite: PolyhedralSiteCollection,
                              VoronoiSite: VoronoiSiteCollection,
                              SphericalSite: SphericalSiteCollection }

    def __init__(self, sites, atoms):
        site_collection_class = None
        for k, v in Trajectory.site_collection_types.items():
            if all( [ isinstance( s, k ) for s in sites ] ):
                site_collection_class = v
        if not site_collection_class:
            raise ValueError('Unable to detect unique SiteCollection class')
        self.sites = sites
        self.atoms = atoms
        self.site_collection = site_collection_class(sites)
        self.timesteps = []
        self.previous_occupations = {}
        self.atom_lookup = {a.index: i for i, a in enumerate(atoms)}
        self.site_lookup = {s.index: i for i, s in enumerate(sites)}

    def atom_by_index(self, i):
        return self.atoms[self.atom_lookup[i]] 

    def site_by_index(self, i):
        return self.sites[self.site_lookup[i]] 

    def analyse_structure(self, structure):
        self.site_collection.analyse_structure(self.atoms, structure)
        
    def assign_site_occupations(self, structure):
        self.site_collection.assign_site_occupations(self.atoms, structure)
                    
    def site_coordination_numbers(self):
        return Counter( [ s.coordination_number for s in self.sites ] )

    def site_labels(self):
        return [ s.label for s in self.sites ]
   
    @property
    def atom_sites(self):
        return [ atom.in_site for atom in self.atoms ]
        
    @property
    def site_occupations(self):
        return [ s.contains_atoms for s in self.sites ]

    def append_timestep(self, structure, t=None):
        self.analyse_structure(structure)
        for atom in self.atoms:
            atom.trajectory.append( atom.in_site )
        for site in self.sites:
            site.trajectory.append( site.contains_atoms )
        self.timesteps.append(t)

    def reset(self):
        for atom in self.atoms:
            atom.reset()
        for site in self.sites:
            site.reset()
        self.timesteps = [] 

    @property
    def atoms_trajectory(self):
        return list(map(list, zip(*[atom.trajectory for atom in self.atoms])))

    @property
    def sites_trajectory(self):
        return list(map(list, zip(*[site.trajectory for site in self.sites])))

    @property
    def at(self):
        return self.atoms_trajectory

    @property
    def st(self):
        return self.sites_trajectory

    def trajectory_from_structures(self, structures, progress=False):
        generator = enumerate( structures, 1 )
        if progress:
            if progress=='notebook':
                generator = tqdm_notebook( generator, total=len(structures), unit=' steps' )
            else:
                generator = tqdm( generator, total=len(structures), unit=' steps' )
        for timestep, s in generator:
            self.append_timestep( s, t=timestep )
   
    def __len__(self):
        """Returns the "length" of a trajectory, i.e. the number of analysed timesteps."""
        return len(self.timesteps)
 
def update_occupation( site, atom ):
    site.contains_atoms.append( atom.index )
    atom.in_site = site.index




class SiteCollection(object):
    """Parent class for collections of sites.

    Collections of specific site types should inherit from this class.

    Attributes:
        sites (list): List of ``Site``-like objects.

    """

    def __init__(self, sites):
        """Create a SiteCollection object.

        Args:
            sites (list): List of ``Site`` objects.

        """
        self.sites = sites

    def assign_site_occupations(self, atoms, structure):
        """Assigns atoms to sites for a specific structure.

        This method should be implemented in the derived subclass

        Args:
            atoms (list(Atom)): List of Atom objects to be assigned to sites.
            struture (pymatgen.Structure): Pymatgen Structure object used to specificy
                the atomic coordinates.

        Returns:
            None

        Notes:
            The atom coordinates should already be consistent with the coordinates
            in `structure`. Recommended usage is via the ``analyse_structure()`` method.

        """
        raise NotImplementedError('assign_site_occupations should be implemented in'
            ' the derived class')

    def analyse_structure(self, atoms, structure):
        """Perform a site analysis for a set of atoms on a specific structure.

        This method should be implemented in the derived subclass.

        Args:
            atoms (list(Atom)): List of Atom objects to be assigned to sites.
            struture (pymatgen.Structure): Pymatgen Structure object used to specificy
                the atomic coordinates.

        Returns:
            None

        """
        raise NotImplementedError('analyse_structure should be implemented in the derived class')

    def neighbouring_sites(self, site_index):
        """If implemented, returns a list of sites that neighbour
        a given site.

        This method should be implemented in the derived subclass.
        
        Args:
            site_index (int): Index of the site to return a list of neighbours for.

        """
        raise NotImplementedError('neighbouring_sites should be implemented'
            'in the derived class')

    def site_by_index(self, index):
        """Returns the site with a specific index.

        Args:
            index (int): index for the site to be returned.

        Returns:
            (Site)

        Raises:
            ValueError: If a site with the specified index is not contained
                in this SiteCollection.

        """
        for site in self.sites:
            if site.index == index:
                return site
        raise ValueError(f'No site with index {index} found')

    def update_occupation(self, site, atom):
        """Updates site and atom attributes for this atom occupying this site.

        Args:
            site (Site): The site to be updated.
            atom (Atom): The atom to be updated.

        Returns:
            None

        Notes:
            This method does the following:
            1. If the atom has moved to a new site, 
               record a old_site --> new_site transition.
            2. Add this atom's index to the list of atoms occupying this site.
            3. Add this atom's fractional coordinates to the list of
               coordinates observed occupying this site.
            4. Assign this atom this site index.

        """
        if atom.in_site:
            if atom.in_site != site.index: # this atom has moved
                previous_site = self.site_by_index(atom.in_site)
                previous_site.transitions[site.index] += 1
        site.contains_atoms.append(atom.index)
        site.points.append(atom.frac_coords)
        atom.in_site = site.index

    def reset_site_occupations(self):
        """Occupations of all sites in this site collection are set as empty.

        Args:
            None

        Returns:
            None

        """
        for s in self.sites:
            s.contains_atoms = []

    def sites_contain_points(self, points):
        """If implemented, Checks whether the set of sites contain 
        a corresponding set of fractional coordinates.

        Args:
            points (np.array): 3xN numpy array of fractional coordinates.
                There should be one coordinate for each site being checked.
        
        Returns:
            (bool)

        Notes:
            Specific SiteCollection subclass implementations may require
            additional arguments to be passed.

        """
        raise NotImplementedError('sites_contain_points() should be'
            ' implemented in the derived class')




from .site import Site

class SphericalSite(Site):
    
    def __init__(self, frac_coords, rcut, label=None):
        super(SphericalSite, self).__init__(label=label)
        self.frac_coords = frac_coords
        self.rcut = rcut

    def as_dict(self):
        d = super(SphericalSite, self).as_dict()
        d['frac_coords'] = self.frac_coords
        d['rcut'] = self.rcut
        return d

    def contains_atom(self, atom, lattice):
        return self.contains_point(atom.frac_coords, lattice)

    def contains_point(self, x, lattice):
        dr = lattice.get_distance_and_image( self.frac_coords, x )[0]
        return dr <= self.rcut

    @classmethod
    def from_dict(cls, d):
        spherical_site = cls( frac_coords=d['frac_coords'],
                              rcut=d['rcut'] )
        spherical_site.label = d.get('label')
        return spherical_site 





from collections import Counter

class Site(object):
    """Parent class for defining sites.

    A Site is a bounded volume that can contain none, one, or more atoms.
    This class defines the attributes and methods expected for specific
    Site subclasses.

    Attributes:
        index (int): Numerical ID, intended to be unique to each site.
        label (`str`: optional): Optional string given as a label for this site.
            Default is `None`.
        contains_atoms (list): List of the atoms contained by this site in the
            structure last processed.
        trajectory (list): List of sites this atom has visited at each timestep?
        points (list): List of fractional coordinates for atoms assigned as
            occupying this site.
        transitions (collections.Counter): Stores observed transitions from this
            site to other sites. Format is {index: count} with ``index`` giving
            the index of each destination site, and ``count`` giving the number 
            of observed transitions to this site.
 
    """

    _newid = 0
    # Site._newid provides a counter that is incremented each time a 
    # new site is initialised. This allows each site to have a 
    # unique numerical index.
    # Site._newid can be reset to 0 by calling Site.reset_index()
    # with the default arguments.
    
    def __init__(self, label=None):
        """Initialise a Site object.

        Args:
            label (`str`: optional): Optional string used to label this site.

        Retuns:
            None

        """
        self.index = Site._newid
        Site._newid += 1
        self.label = label
        self.contains_atoms = []
        self.trajectory = []
        self.points = []
        self.transitions = Counter()

    def reset(self):
        """Reset the trajectory for this site.

        Returns the contains_atoms and trajectory attributes
        to empty lists.

        Args:
            None

        Returns:
            None

        """
        self.contains_atoms = []
        self.trajectory = []
        self.transitions = Counter()
 
    def contains_point(self, x):
        """Test whether the fractional coordinate x is contained by this site.

        This method should be implemented in the inherited subclass

        Args:
            x (np.array): Fractional coordinate.

        Returns:
            (bool)

        Note:
            Specific Site subclasses may require additional arguments to be passed.

        """
        raise NotImplementedError('contains_point should be implemented '
                                  'in the inherited class')

    def contains_atom(self, atom):
        """Test whether this site contains a specific atom.

        Args:
            atom (Atom): The atom to test.

        Returns:
            (bool)

        """
        return self.contains_point(atom.frac_coords)

    def as_dict(self):
        """Json-serializable dict representation of this Site.

        Args:
            None

        Returns:
            (dict)

        """
        d = {'index': self.index,
             'contains_atoms': self.contains_atoms,
             'trajectory': self.trajectory,
             'points': self.points,
             'transitions': self.transitions}
        if self.label:
            d['label'] = self.label
        return d

    @classmethod
    def from_dict(cls, d):
        """Create a Site object from a dict representation.

        Args:
            d (dict): The dict representation of this Site.

        Returns:
            (Site)

        """
        site = cls()
        site.index = d['index']
        site.trajectory = d['trajectory']
        site.contains_atoms = d['contains_atoms']
        site.points = d['points']
        site.transitions = d['transitions']
        site.label = d.get('label')
        return site 

    def centre(self):
        """Returns the centre point of this site.

        This method should be implemented in the inherited subclass.

        Args:
            None

        Returns:
            None

        """ 
        raise NotImplementedError('centre should be implemeneted '
                                  'in the inherited class')

    @classmethod
    def reset_index(cls, newid=0):
        """Reset the site index counter.

        Args:
            newid (`int`: optional): New starting index. Default is 1.

        Returns:
            None

        """ 
        Site._newid = newid



from .site_collection import SiteCollection
import numpy as np

class VoronoiSiteCollection(SiteCollection):

    def analyse_structure(self, atoms, structure):
        for a in atoms:
            a.assign_coords(structure)
        self.assign_site_occupations(atoms, structure)

    def assign_site_occupations(self, atoms, structure):
        self.reset_site_occupations()
        lattice = structure.lattice
        site_coords = np.array( [ s.frac_coords for s in self.sites ] )
        atom_coords = np.array( [ a.frac_coords for a in atoms ] )
        dist_matrix = lattice.get_all_distances(site_coords, atom_coords)
        site_list_indices = np.argmin( dist_matrix, axis=0 )
        for atom, site_list_index in zip( atoms, site_list_indices):
            site = self.sites[site_list_index]
            self.update_occupation( site, atom )

    
 



import numpy as np

def get_nearest_neighbour_indices(structure, ref_structure, vertex_species, n_coord):
    """
    Returns the atom indices for the N nearest neighbours to each site in a reference
    structure.

    Args:
        structure (`pymatgen.Structure`): A pymatgen Structure object, used to select
            the nearest neighbour indices.
        ref_structure (`pymatgen.Structure`): A pymatgen Structure object. Each site
            is used to find the set of N nearest neighbours (of the specified atomic species)
            in ``structure``.
        vertex_species (list(str)): List of strings specifying the atomic species of
            the vertex atoms, e.g. ``[ 'S', 'I' ]``.
        n_coord (int): Number of matching nearest neighbours to return for each site in 
            ``ref_structure``.

    Returns:
        (list(list(int)): N_sites x N_neighbours nested list of vertex atom indices.

    """
    vertex_indices = [ i for i, s in enumerate(structure) if s.species_string in vertex_species ]
    struc1_coords = np.array([structure[i].frac_coords for i in vertex_indices])
    struc2_coords = ref_structure.frac_coords
    lattice = structure[0].lattice
    dr_ij = lattice.get_all_distances(struc1_coords, struc2_coords).T
    nn_indices = []
    for dr_i in dr_ij:
        idx = np.argpartition(dr_i, n_coord)
        nn_indices.append( sorted([ vertex_indices[i] for i in idx[:n_coord] ]) )
    return nn_indices

def get_vertex_indices( structure, centre_species, vertex_species, cutoff=4.5, n_vertices=6 ):
    """
    Find the atom indices for atoms defining the vertices of coordination polyhedra, from 
    a pymatgen Structure object.

    Given the elemental species of a set of central atoms, A, 
    and of the polyhedral vertices, B, this function finds:
    for each A, then N closest neighbours B (within some cutoff).
    The number of neighbours found per central atom can be a single
    value for all A, or can be provided as a list of values for each A.

    Args:
        structure (`pymatgen.Structure`): A pymatgen Structure object, used to
            find the coordination polyhedra vertices..
        centre_species (str): Species string identifying the atoms at the centres
            of each coordination environment, e.g. "Na".
        vertex_species (str or list(str)): Species string identifying the atoms at the vertices
            of each coordination environment, e.g. "S"., or a list of strings, e.g. ``["S", "I"]``.
        cutoff (float): Distance cutoff for neighbour search.
        n_vertices (int or list(int)): Number(s) of nearest neighbours to return
            for each set of vertices. If a list is passed, this should be the same
            length as the number of atoms of centre species A.

    Returns:
        list(list(int)): Nested list of integers, giving the atom indices for each
            coordination environment.

    """
    central_sites = [ s for s in structure if s.species_string == centre_species ]
    if isinstance(n_vertices, int):
        n_vertices = [n_vertices] * len(central_sites)
    if isinstance(vertex_species, str):
        vertex_species = [ vertex_species ]
    vertex_indices = []
    for site, n_vert in zip(central_sites, n_vertices):
        neighbours = [ s for s in structure.get_neighbors(site, r=cutoff, include_index=True) 
                       if s[0].species_string in vertex_species ]
        neighbours.sort(key=lambda x: x[1])
        atom_indices = [ n[2] for n in neighbours[:n_vert] ]
        vertex_indices.append( atom_indices )
    return vertex_indices

def x_pbc(x):
    all_x =  np.array([[0,0,0],
                       [1,0,0],
                       [0,1,0],
                       [0,0,1],
                       [1,1,0],
                       [1,0,1],
                       [0,1,1],
                       [1,1,1]]) + x
    return all_x

def species_string_from_site(site):
    return [k.__str__() for k in site._species.keys()][0]



from .site_collection import SiteCollection

class PolyhedralSiteCollection(SiteCollection):
    """A collection of PolyhedralSite objects.

    Attributes:
        sites (list): List of ``Site``-like objects.

    """

    def __init__(self, sites):
        """Create a PolyhedralSiteCollection instance.

        Args:
            sites (list(PolyhedralSite)): List of PolyhedralSite objects.

        Returns:
            None

        """
        super(PolyhedralSiteCollection, self).__init__(sites)
        self._neighbouring_sites = construct_neighbouring_sites(self.sites)

    def analyse_structure(self, atoms, structure):
        for a in atoms:
            a.assign_coords(structure)
        for s in self.sites:
            s.assign_vertex_coords(structure)
        self.assign_site_occupations(atoms, structure)

    def assign_site_occupations(self, atoms, structure):
        self.reset_site_occupations()
        for atom in atoms:
            if atom.in_site:
                # first check the site last occupied
                previous_site = next(s for s in self.sites if s.index == atom.in_site)
                if previous_site.contains_atom(atom):
                    self.update_occupation( previous_site, atom )
                    continue # atom has not moved
                else: # default is atom does not occupy any sites
                    atom.in_site = None
            for s in self.sites:
                if s.contains_atom(atom):
                    self.update_occupation( s, atom )
                    break

    def neighbouring_sites(self, index):
        return self._neighbouring_sites[index] 

    def sites_contain_points(self, points, structure):
        """Checks whether the set of sites contain 
        a corresponding set of fractional coordinates.

        Args:
            points (np.array): 3xN numpy array of fractional coordinates.
                There should be one coordinate for each site being checked.
            structure (Structure): Pymatgen Structure used to define the
                vertex coordinates of each polyhedral site.
        
        Returns:
            (bool)

        """
        check = all([s.contains_point(p,structure) for s, p in zip(self.sites, points)])
        return check

def construct_neighbouring_sites(sites):
    """
    Find all polyhedral sites that are face-sharing neighbours.

    Any polyhedral sites that share 3 or more vertices are considered
    to share a face.

    Args:
        None

    Returns:
        (dict): Dictionary of `int`: `list` entries. 
            Keys are site indices. Values are lists of ``PolyhedralSite`` objects.

    """
    neighbours = {}
    for site_i in sites:
        neighbours[site_i.index] = []
        for site_j in sites:
            if site_i is site_j:
                continue
            # 3 or more common vertices indicated a shared face.
            n_shared_vertices = len(set(site_i.vertex_indices) & set(site_j.vertex_indices))
            if n_shared_vertices >= 3:
                neighbours[site_i.index].append(site_j)
    return neighbours
 



from setuptools import setup, find_packages
from site_analysis import __version__ as VERSION

readme = 'README.md'
long_description = open( readme ).read()

config = {
    'description': 'MD analysis using site occupation trajectories',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'author': 'Benjamin J. Morgan',
    'author_email': 'b.j.morgan@bath.ac.uk',
    'url': 'https://github.com/bjmorgan/site-analysis',
    'download_url': "https://github.com/bjmorgan/site-analysis/archive/%s.tar.gz" % (VERSION),
    'author_email': 'b.j.morgan@bath.ac.uk',
    'version': VERSION,
    'install_requires': open( 'requirements.txt' ).read(),
    'python_requires': '>=3.6',
    'license': 'MIT',
    'packages': [ 'site_analysis' ],
    'scripts': [],
    'name': 'site-analysis'
}

setup(**config)



from .site_collection import SiteCollection
import numpy as np

class SphericalSiteCollection(SiteCollection):


    def analyse_structure(self, atoms, structure):
        for a in atoms:
            a.get_coords(structure)
        self.assign_site_occupations(atoms, structure)

    def assign_site_occupations(self, atoms, structure):
        self.reset_site_occupations()
        for atom in atoms:
            if atom.in_site:
                # first check the site last occupied
                previous_site = next(s for s in self.sites if s.index == atom.in_site)
                if previous_site.contains_atom(atom, structure.lattice):
                    self.update_occupation( previous_site, atom )
                    continue # atom has not moved
                else: # default is atom does not occupy any sites
                    atom.in_site = None
            for s in self.sites:
                if s.contains_atom(atom, structure.lattice):
                    self.update_occupation( s, atom )
                    break

 



import itertools
import numpy as np 
from scipy.spatial import Delaunay, ConvexHull
from .site import Site
from .tools import x_pbc, species_string_from_site

class PolyhedralSite(Site):
    """Describes a site defined by the polyhedral volume enclosed by a set
    of vertex atoms.

    Attributes:
        index (int): Numerical ID, intended to be unique to each site.
        label (`str`: optional): Optional string given as a label for this site.
            Default is `None`.
        contains_atoms (list): List of the atoms contained by this site in the
            structure last processed.
        trajectory (list): List of sites this atom has visited at each timestep?
        points (list): List of fractional coordinates for atoms assigned as
            occupying this site.
        transitions (collections.Counter): Stores observed transitions from this
            site to other sites. Format is {index: count} with ``index`` giving
            the index of each destination site, and ``count`` giving the number 
            of observed transitions to this site.
        vertex_indices (list(int)): List of integer indices for the vertex atoms
            (counting from 0). 
        label (:obj:`str`, optional): Optional label for the site.
   
    """ 

    def __init__(self, vertex_indices, label=None):
        """Create a PolyhedralSite instance.

        Args:
            vertex_indices (list(int)): List of integer indices for the vertex atoms (counting from 0).
            label (:obj:`str`, optional): Optional label for this site.

        Returns:
            None

        """
        super(PolyhedralSite, self).__init__(label=label)
        self.vertex_indices = vertex_indices
        self.vertex_coords = None
        self._delaunay = None

    def __repr__(self):
        string = ('site_analysis.PolyhedralSite('
                  f'index={self.index}, '
                  f'label={self.label}, '
                  f'vertex_indices={self.vertex_indices}, '
                  f'contains_atoms={self.contains_atoms})')
        return string
                  
    def reset(self):
        """Reset the trajectory for this site.

        Resets the contains_atoms and trajectory attributes
        to empty lists.

        Vertex coordinates and Delaunay tesselation are unset.

        Args:
            None

        Returns:
            None

        """
        super(PolyhedralSite, self).reset()
        self.vertex_coords = None
        self._delaunay = None
 
    @property
    def delaunay(self):
        """Delaunay tesselation of the vertex coordinates for this site.

        This is calculated the first time the attribute is requested,
        and then stored for reuse, unless the site is reset.

        Returns:
            (scipy.spatial.Delaunay)

        """
        if not self._delaunay:
            self._delaunay = Delaunay(self.vertex_coords)         
        return self._delaunay

    @property
    def coordination_number(self):
        """Coordination number for this site, defined as the number of 
        vertices

        Returns:
            (int)

        """
        return len(self.vertex_indices)
    
    @property
    def cn(self):
        """Coordination number for this site, defined as the number of
        vertices

        Convenience property for coordination_number()

        Returns:
            (int)

        """
        return self.coordination_number
        
    def assign_vertex_coords(self, structure):
        """Assign fractional coordinates to the polyhedra vertices
        from the corresponding atom positions in a pymatgen Structure.

        Args:
            structure (Structure): The pymatgen Structure used to assign
                the vertices fractional coordinates.

        Returns:
            None

        Notes:
            This method assumes the coordinates of the vertices may 
            have changed, so unsets the Delaunay tesselation for this site.

        """
        frac_coords = np.array([ s.frac_coords for s in 
            [ structure[i] for i in self.vertex_indices ] ] )
        # Handle periodic boundary conditions:
        # If the range of fractional coordinates along x, y, or z 
        # exceeds 0.5, assume that this polyhedron wraps around the 
        # periodic boundary in that dimension. 
        # Fractional coordinates for that dimension that are less 
        # than 0.5 will be incremented by 1.0 
        for i in range(3):
            spread = max(frac_coords[:,i]) - min(frac_coords[:,i])
            if spread > 0.5:
                for j, fc in enumerate(frac_coords):
                    if fc[i] < 0.5:
                        frac_coords[j,i] += 1.0
        self.vertex_coords = frac_coords
        self._delaunay = None

    def get_vertex_species(self, structure):
        """Returns a list of species strings for the vertex atoms of this
        polyhedral site.

        Args:
            structure (Structure): Pymatgen Structure used to assign species
                to each vertex atom.

        Returns:
            (list(str)): List of species strings of the vertex atoms.

        """
        return [structure[i].species_string for i in self.vertex_indices]

    def contains_point(self, x, structure=None, algo='simplex'):
        """Test whether a specific point is enclosed by this polyhedral site.

        Args:
            x (np.array): Fractional coordinates of the point to test (length 3 numpy array).
            structure (:obj:`Structure`, optional): Optional pymatgen Structure. If provided,
                the vertex coordinates for this polyhedral site will be assigned using
                this structure. Default is None.
            algo (str): Select the algorithm for testing whether a point is contained
                by the site:
    
                simplex: Use scipy.spatial.Delaunay.find_simplex to test if any of
                         the simplices that make up this polyhedron contain the point.
       
                sn:      Compute the sign of the surface normal for each polyhedron 
                         face with respect to the point, to test if the point lies
                         "inside" every face.
                
        Returns:
            (bool)

        """
        contains_point_algos = {'simplex': self.contains_point_simplex,
                                'sn': self.contains_point_sn}
        if algo not in contains_point_algos.keys():
            raise ValueError(f'{algo} is not a valid algorithm keyword for contains_point()')
        if structure:
            self.assign_vertex_coords(structure)
        if self.vertex_coords is None:
            raise RuntimeError('no vertex coordinates set for polyhedral_site {}'.format(self.index))
        return contains_point_algos[algo](x_pbc(x))
   
    def contains_point_simplex(self, x):
        """Test whether one or more points are inside this site, by checking 
        whether these points are contained inside the simplices of the Delaunay 
        tesselation defined by the vertex coordinates.

        Args:
            x (np.array): Fractional coordinates for one or more points, as a
                (3x1) or (3xN) numpy array.

        Returns:
            (bool)

        """
        return np.any(self.delaunay.find_simplex(x) >= 0)
 
    def contains_point_sn(self, x_list):
        """Test whether one or more points are inside this site, by calculating 
        the sign of the surface normal for each face with respect to each point.

        Args:
            x (np.array): Fractional coordinates for one or more points, as a
                (3x1) or (3xN) numpy array.

        Returns:
            (bool)

        Note:
            This method could be made more efficient by caching the 
            surface_normal vectors and in-face vectors.

            This is also a possible target for optimisation with f2py etc.

        """
        hull = ConvexHull(self.vertex_coords)
        faces = hull.points[hull.simplices]
        centre = self.centre()
        inside = []
        for x in x_list:
            dotsum = 0
            for f in faces:
                surface_normal = np.cross(f[0]-f[2], f[1]-f[2])
                c_sign = np.sign(np.dot( surface_normal, centre-f[0]))
                p_sign = np.sign(np.dot( surface_normal, x-f[0]))
                dotsum += c_sign * p_sign
                inside.append(dotsum == len(faces))
        return any(inside)

    def contains_atom(self, atom, algo='simplex'):
        """Test whether an atom is inside this polyhedron.

        Args:
            atom (Atom): The atom to test.
            algo (:obj:`str`, optional): Select the algorithm to us. Options are
                'simplex' and 'sn'. See the documentation for the contains_point()
                method for more details. Default is 'simplex'.

        Returns:
            (bool)
        """
        contains_point_algos = ['simplex', 'sn']
        if algo not in contains_point_algos:
            raise ValueError(f'{algo} is not a valid algorithm keyword for contains_atom()')
        return self.contains_point(atom.frac_coords, algo=algo)

    def as_dict(self):
        d = super(PolyhedralSite, self).as_dict()
        d['vertex_indices'] = self.vertex_indices
        d['vertex_coords'] = self.vertex_coords
        return d

    @classmethod
    def from_dict(cls, d):
        polyhedral_site = cls(vertex_indices=d['vertex_indices'])
        polyhedral_site.vertex_coords = d['vertex_coords']
        polyhedral_site.contains_atoms = d['contains_atoms']
        polyhedral_site.label = d.get('label')
        return polyhedral_site 

    def centre(self):
        """Returns the fractional coordinates of the centre point of
        this polyhedral site.

        Args:
            None

        Returns:
            (np.array): (3,) numpy array.
 
        """
        return np.mean(self.vertex_coords, axis=0)

    @classmethod
    def sites_from_vertex_indices(cls, vertex_indices, label=None):
        sites = [cls(vertex_indices=vi, label=label) for vi in vertex_indices]
        return sites
